import torch.nn as nn
import torch
import math
import torch.utils.model_zoo as model_zoo
from torchvision.ops import nms
<<<<<<< HEAD
from .utils import BasicBlock, Bottleneck, BBoxTransform, ClipBoxes
from .anchors import Anchors
from .losses import calc_iou, FocalLoss
from .encoders import SeResNetXtEncoder, ResNetEncoder
=======
from utils import BasicBlock, Bottleneck, BBoxTransform, ClipBoxes
from anchors import Anchors
import losses
from encoders import SeResNetXtEncoder, ResNetEncoder
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


class PyramidFeatures(nn.Module):
    def __init__(self, C2_size, C3_size, C4_size, C5_size, feature_size=256, use_l2_features=True):
        super(PyramidFeatures, self).__init__()
        self.use_l2_features = use_l2_features

        # upsample C5 to get P5 from the FPN paper
        self.P5_1           = nn.Conv2d(C5_size, feature_size, kernel_size=1, stride=1, padding=0)
        self.P5_upsampled   = nn.Upsample(scale_factor=2, mode='nearest')
        self.P5_2           = nn.Conv2d(feature_size, feature_size, kernel_size=3, stride=1, padding=1)

        # add P5 elementwise to C4
        self.P4_1           = nn.Conv2d(C4_size, feature_size, kernel_size=1, stride=1, padding=0)
        self.P4_upsampled   = nn.Upsample(scale_factor=2, mode='nearest')
        self.P4_2           = nn.Conv2d(feature_size, feature_size, kernel_size=3, stride=1, padding=1)

        # add P4 elementwise to C3
        self.P3_1 = nn.Conv2d(C3_size, feature_size, kernel_size=1, stride=1, padding=0)
        self.P3_upsampled = nn.Upsample(scale_factor=2, mode='nearest')
        self.P3_2 = nn.Conv2d(feature_size, feature_size, kernel_size=3, stride=1, padding=1)

        # add P3 elementwise to C2
        self.P2_1 = nn.Conv2d(C2_size, feature_size, kernel_size=1, stride=1, padding=0)
        self.P2_2 = nn.Conv2d(feature_size, feature_size, kernel_size=3, stride=1, padding=1)

        # "P6 is obtained via a 3x3 stride-2 conv on C5"
        self.P6 = nn.Conv2d(C5_size, feature_size, kernel_size=3, stride=2, padding=1)

        # "P7 is computed by applying ReLU followed by a 3x3 stride-2 conv on P6"
        self.P7_1 = nn.ReLU()
        self.P7_2 = nn.Conv2d(feature_size, feature_size, kernel_size=3, stride=2, padding=1)

    def forward(self, inputs):
        C2, C3, C4, C5 = inputs

        P5_x = self.P5_1(C5)
        P5_upsampled_x = self.P5_upsampled(P5_x)
        P5_x = self.P5_2(P5_x)
        
        P4_x = self.P4_1(C4)
        P4_x = P5_upsampled_x + P4_x
        P4_upsampled_x = self.P4_upsampled(P4_x)
        P4_x = self.P4_2(P4_x)

        P3_x = self.P3_1(C3)
        P3_x = P3_x + P4_upsampled_x
        P3_upsampled_x = self.P3_upsampled(P3_x)
        P3_x = self.P3_2(P3_x)

        if self.use_l2_features:
            P2_x = self.P2_1(C2)
            P2_x = P2_x + P3_upsampled_x
            P2_x = self.P2_2(P2_x)

        P6_x = self.P6(C5)

        P7_x = self.P7_1(P6_x)
        P7_x = self.P7_2(P7_x)

        if self.use_l2_features:
            return [P2_x, P3_x, P4_x, P5_x, P6_x, P7_x]
        else:
            return [P3_x, P4_x, P5_x, P6_x, P7_x]


class RegressionModel(nn.Module):
    def __init__(self, num_features_in, num_anchors=9, feature_size=256):
        super(RegressionModel, self).__init__()

        self.conv1 = nn.Conv2d(num_features_in, feature_size, kernel_size=3, padding=1)
        self.act1 = nn.ReLU()

        self.conv2 = nn.Conv2d(feature_size, feature_size, kernel_size=3, padding=1)
        self.act2 = nn.ReLU()

        self.conv3 = nn.Conv2d(feature_size, feature_size, kernel_size=3, padding=1)
        self.act3 = nn.ReLU()

        self.conv4 = nn.Conv2d(feature_size, feature_size, kernel_size=3, padding=1)
        self.act4 = nn.ReLU()

        self.output = nn.Conv2d(feature_size, num_anchors * 4, kernel_size=3, padding=1)

    def forward(self, x):
        out = self.conv1(x)
        out = self.act1(out)

        out = self.conv2(out)
        out = self.act2(out)

        out = self.conv3(out)
        out = self.act3(out)

        out = self.conv4(out)
        out = self.act4(out)

        out = self.output(out)

        # out is B x C x W x H, with C = 4*num_anchors
        out = out.permute(0, 2, 3, 1)

        return out.contiguous().view(out.shape[0], -1, 4)


class ClassificationModel(nn.Module):
    def __init__(self, num_features_in, num_anchors=9, num_classes=80, prior=0.01, feature_size=256, dropout=0.5):
        super(ClassificationModel, self).__init__()

        self.num_classes = num_classes
        self.num_anchors = num_anchors

        self.conv1 = nn.Conv2d(num_features_in, feature_size, kernel_size=3, padding=1)
        self.act1 = nn.ReLU()

        self.conv2 = nn.Conv2d(feature_size, feature_size, kernel_size=3, padding=1)
        self.act2 = nn.ReLU()

        self.conv3 = nn.Conv2d(feature_size, feature_size, kernel_size=3, padding=1)
        self.act3 = nn.ReLU()

        self.conv4 = nn.Conv2d(feature_size, feature_size, kernel_size=3, padding=1)
        self.act4 = nn.ReLU()

        self.output = nn.Conv2d(feature_size, num_anchors * num_classes, kernel_size=3, padding=1)
        self.output_act = nn.Sigmoid()

        self.dropout = dropout

    def forward(self, x):
        out = self.conv1(x)
        out = self.act1(out)

        out = self.conv2(out)
        out = self.act2(out)

        out = self.conv3(out)
        out = self.act3(out)

        out = self.conv4(out)
        out = self.act4(out)

        if self.dropout > 0:
            out = nn.functional.dropout(out, self.dropout, self.training)

        out = self.output(out)
        out = self.output_act(out)

        # out is B x C x W x H, with C = n_classes + n_anchors
        out1 = out.permute(0, 2, 3, 1)

        batch_size, width, height, channels = out1.shape

        out2 = out1.view(batch_size, width, height, self.num_anchors, self.num_classes)

        return out2.contiguous().view(x.shape[0], -1, self.num_classes)

class GlobalClassificationModel(nn.Module):
    def __init__(self, num_features_in, num_classes=80, feature_size=256, dropout=0.5):
        super().__init__()

        self.num_classes = num_classes
        self.conv1 = nn.Conv2d(num_features_in, feature_size, kernel_size=3, dilation=1, padding=0)
        self.fc = nn.Linear(feature_size*2, num_classes)
        self.output_act = nn.LogSoftmax(dim=-1)
<<<<<<< HEAD

        self.dropout = dropout

    def forward(self, x):
        out = nn.functional.max_pool2d(x, 2)
        out = self.conv1(out)
        out = nn.functional.relu(out)

        # if self.dropout > 0:
        #     out = F.dropout(out, self.dropout, self.training)

        avg_pool = nn.functional.avg_pool2d(out, out.shape[2:])
        max_pool = nn.functional.max_pool2d(out, out.shape[2:])
        avg_max_pool = torch.cat((avg_pool, max_pool), 1)
        out = avg_max_pool.view(avg_max_pool.size(0), -1)

        if self.dropout > 0:
            out = nn.functional.dropout(out, self.dropout, self.training)

        out = self.fc(out)
        out = self.output_act(out)

        return out

=======

        self.dropout = dropout

    def forward(self, x):
        out = F.max_pool2d(x, 2)
        out = self.conv1(out)
        out = F.relu(out)

        # if self.dropout > 0:
        #     out = F.dropout(out, self.dropout, self.training)

        avg_pool = F.avg_pool2d(out, out.shape[2:])
        max_pool = F.max_pool2d(out, out.shape[2:])
        avg_max_pool = torch.cat((avg_pool, max_pool), 1)
        out = avg_max_pool.view(avg_max_pool.size(0), -1)

        if self.dropout > 0:
            out = F.dropout(out, self.dropout, self.training)

        out = self.fc(out)
        out = self.output_act(out)

        return out

>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640
class RetinaNetEncoder(nn.Module):
    def __init__(self):
        super().__init__()
        self.fpn_sizes = []

    def forward(self, x):
        """
        :param x: input tensor
        :return: x1, x2, x3, x4 layer outputs
        """
        raise NotImplementedError()

class RetinaNet(nn.Module):

<<<<<<< HEAD
    def __init__(self, encoder: RetinaNetEncoder, num_classes, dropout_cls=0.5, dropout_global_cls=0.5, use_l2_features=True):
        self.inplanes = 64
        super(RetinaNet, self).__init__()

        print('Dropout_cls:', dropout_cls, '   Dropout_global_cls:', dropout_global_cls)

        fpn_sizes = encoder.fpn_sizes
        self.use_l2_features = use_l2_features

        self.fpn = PyramidFeatures(fpn_sizes[0], fpn_sizes[1], fpn_sizes[2], fpn_sizes[3], use_l2_features=use_l2_features)
=======
    def __init__(self, encoder: RetinaNetEncoder, num_classes, block, layers, dropout_cls=0.5, dropout_global_cls=0.5, use_l2_features=True):
        self.inplanes = 64
        super(RetinaNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)

        fpn_sizes = encoder.fpn_sizes
        self.use_l2_features = use_l2_features

        self.fpn = PyramidFeatures(fpn_sizes[0], fpn_sizes[1], fpn_sizes[2], use_l2_features=use_l2_features)
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640

        self.regressionModel = RegressionModel(256)
        self.classificationModel = ClassificationModel(256, num_classes=num_classes, dropout=dropout_cls)

<<<<<<< HEAD
        self.globalClassificationModel = GlobalClassificationModel(fpn_sizes[-1], num_classes=num_classes, feature_size=256, dropout=dropout_global_cls)
        self.globalClassificationLoss = nn.NLLLoss()

        if use_l2_features:
            pyramid_levels = [2, 3, 4, 5, 6, 7]
        else:
            pyramid_levels = [3, 4, 5, 6, 7]
=======
        self.globalClassificationModel = GlobalClassificationModel(fpn_sizes[-1], num_classes=3, feature_size=256, dropout=dropout_global_cls)
        self.globalClassificationLoss = nn.NLLLoss()
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640

        self.anchors = Anchors(pyramid_levels=pyramid_levels)

        self.regressBoxes = BBoxTransform()

        self.clipBoxes = ClipBoxes()

        self.focalLoss = FocalLoss()

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

        self.encoder = encoder

        prior = 0.01

        self.classificationModel.output.weight.data.fill_(0)
        self.classificationModel.output.bias.data.fill_(-math.log((1.0 - prior) / prior))

        self.regressionModel.output.weight.data.fill_(0)
        self.regressionModel.output.bias.data.fill_(0)

        self.freeze_bn()

    def freeze_bn(self):
        '''Freeze BatchNorm layers.'''
        for layer in self.modules():
            if isinstance(layer, nn.BatchNorm2d):
                layer.eval()
            
    def freeze_encoder(self):
        # self.encoder.eval()
        for param in self.encoder.parameters():
            param.requires_grad = False

    def unfreeze_encoder(self):
        for param in self.encoder.parameters():
            param.requires_grad = True

    def boxes(self, img_batch, regression, classification, global_classification, anchors):
        transformed_anchors = self.regressBoxes(anchors, regression)
        transformed_anchors = self.clipBoxes(transformed_anchors, img_batch)

        finalResult = [[], [], []]

        finalScores = torch.Tensor([])
        finalAnchorBoxesIndexes = torch.Tensor([]).long()
        finalAnchorBoxesCoordinates = torch.Tensor([])

        if torch.cuda.is_available():
            finalScores = finalScores.cuda()
            finalAnchorBoxesIndexes = finalAnchorBoxesIndexes.cuda()
            finalAnchorBoxesCoordinates = finalAnchorBoxesCoordinates.cuda()

        for i in range(classification.shape[2]):
            scores = torch.squeeze(classification[:, :, i])
            scores_over_thresh = (scores > 0.025)
            if scores_over_thresh.sum() == 0:
                # no boxes to NMS, just continue
                continue

            scores = scores[scores_over_thresh]
            anchorBoxes = torch.squeeze(transformed_anchors)
            anchorBoxes = anchorBoxes[scores_over_thresh]
            anchors_nms_idx = nms(anchorBoxes, scores, 0.5)

            finalResult[0].extend(scores[anchors_nms_idx])
            finalResult[1].extend(torch.tensor([i] * anchors_nms_idx.shape[0]))
            finalResult[2].extend(anchorBoxes[anchors_nms_idx])

            finalScores = torch.cat((finalScores, scores[anchors_nms_idx]))
            finalAnchorBoxesIndexesValue = torch.tensor([i] * anchors_nms_idx.shape[0])
            if torch.cuda.is_available():
                finalAnchorBoxesIndexesValue = finalAnchorBoxesIndexesValue.cuda()
<<<<<<< HEAD
=======

            finalAnchorBoxesIndexes = torch.cat((finalAnchorBoxesIndexes, finalAnchorBoxesIndexesValue))
            finalAnchorBoxesCoordinates = torch.cat((finalAnchorBoxesCoordinates, anchorBoxes[anchors_nms_idx]))

        return [finalScores, finalAnchorBoxesIndexes, finalAnchorBoxesCoordinates]
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640

            finalAnchorBoxesIndexes = torch.cat((finalAnchorBoxesIndexes, finalAnchorBoxesIndexesValue))
            finalAnchorBoxesCoordinates = torch.cat((finalAnchorBoxesCoordinates, anchorBoxes[anchors_nms_idx]))

        return [finalScores, finalAnchorBoxesIndexes, finalAnchorBoxesCoordinates]

<<<<<<< HEAD
    def forward(self, inputs, return_loss, return_boxes, return_raw=False):

        if return_loss:
=======
        if self.training:
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640
            img_batch, annotations, global_annotations = inputs
        else:
            img_batch = inputs

        # If working with a 1 channel image:
        # x = torch.cat([img_batch, img_batch, img_batch], dim=1)

        x1, x2, x3, x4 = self.encoder.forward(img_batch)

        features = self.fpn([x1, x2, x3, x4])

        regression = torch.cat([self.regressionModel(feature) for feature in features], dim=1)

        classification = torch.cat([self.classificationModel(feature) for feature in features], dim=1)

        global_classification = self.globalClassificationModel(x4)
<<<<<<< HEAD

        anchors = self.anchors(img_batch)

        if return_raw:
            return [regression, classification, torch.exp(global_classification), anchors]

        result = []

        if return_loss:
            result += list(self.focalLoss(classification, regression, anchors, annotations))
            result += [self.globalClassificationLoss(global_classification, global_annotations)]

        if return_boxes:
=======

        anchors = self.anchors(img_batch)

        result = []

        if self.training:
            result += self.focalLoss(classification, regression, anchors, annotations)
            result += [self.globalClassificationLoss(global_classification, global_annotations)]

        else:
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640
            result += self.boxes(img_batch, regression, classification, global_classification, anchors)

        return result


def resnet18(num_classes, pretrained=False, **kwargs):
    """Constructs a ResNet-18 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    encoder = ResNetEncoder(BasicBlock, [2, 2, 2, 2])
    if pretrained:
<<<<<<< HEAD
        encoder.load_state_dict(model_zoo.load_url(model_urls['resnet18'], model_dir='.'), strict=False)
=======
        model.load_state_dict(model_zoo.load_url(model_urls['resnet18'], model_dir='.'), strict=False)
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640
    model = RetinaNet(encoder=encoder, num_classes=num_classes)
    return model


def resnet34(num_classes, pretrained=False, **kwargs):
    """Constructs a ResNet-34 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    encoder = ResNetEncoder(BasicBlock, [3, 4, 6, 3])
    if pretrained:
<<<<<<< HEAD
        encoder.load_state_dict(model_zoo.load_url(model_urls['resnet34'], model_dir='.'), strict=False)
=======
        model.load_state_dict(model_zoo.load_url(model_urls['resnet34'], model_dir='.'), strict=False)
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640
    model = RetinaNet(encoder=encoder, num_classes=num_classes)
    return model


def resnet50(num_classes, pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    encoder = ResNetEncoder(Bottleneck, [3, 4, 6, 3])
    if pretrained:
<<<<<<< HEAD
        encoder.load_state_dict(model_zoo.load_url(model_urls['resnet50'], model_dir='.'), strict=False)
=======
        model.load_state_dict(model_zoo.load_url(model_urls['resnet50'], model_dir='.'), strict=False)
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640
    model = RetinaNet(encoder=encoder, num_classes=num_classes, **kwargs)
    return model


def resnet101(num_classes, pretrained=False, **kwargs):
    """Constructs a ResNet-101 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    encoder = ResNetEncoder(Bottleneck, [3, 4, 23, 3])
    if pretrained:
<<<<<<< HEAD
        encoder.load_state_dict(model_zoo.load_url(model_urls['resnet101'], model_dir='.'), strict=False)
=======
        model.load_state_dict(model_zoo.load_url(model_urls['resnet101'], model_dir='.'), strict=False)
>>>>>>> cf129a0aec86474c0301cb1640bd781f2d58f640
    model = RetinaNet(encoder=encoder, num_classes=num_classes, **kwargs)
    return model


def resnet152(num_classes, pretrained=False, **kwargs):
    """Constructs a ResNet-152 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    encoder = ResNetEncoder(Bottleneck, [3, 8, 36, 3])
    if pretrained:
        encoder.load_state_dict(model_zoo.load_url(model_urls['resnet152'], model_dir='.'), strict=False)

    model = RetinaNet(encoder=encoder, num_classes=num_classes, **kwargs)
    return model

def se_resnext101(num_classes, pretrained=False, dropout=0.5, fold=0):
    """Constructs a ResNet-101 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    encoder = SeResNetXtEncoder(layers=[3, 4, 23, 3])
    if pretrained == 'imagenet':
        encoder.load_state_dict(model_zoo.load_url(
            senet.pretrained_settings['se_resnext101_32x4d']['imagenet']['url'], model_dir='.'), strict=False)

    model = RetinaNet(encoder=encoder, num_classes=num_classes, dropout_cls=dropout, dropout_global_cls=dropout)
    return model


def se_resnext50(num_classes, pretrained=False, dropout=0.5):
    """Constructs a ResNet-101 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    encoder = SeResNetXtEncoder(layers=[3, 4, 6, 3])
    if pretrained == 'imagenet':
        encoder.load_state_dict(model_zoo.load_url(
            senet.pretrained_settings['se_resnext50_32x4d']['imagenet']['url'], model_dir='.'), strict=False)

    model = RetinaNet(encoder=encoder, num_classes=num_classes, dropout_cls=dropout, dropout_global_cls=dropout)
    return model